package com.wechi.adweb.bridge.google.analytics.controller;

import com.wechi.adweb.bridge.common.APIStatus;
import com.wechi.adweb.bridge.common.BaseController;
import com.wechi.adweb.bridge.common.OpenAPIRequest;
import com.wechi.adweb.bridge.common.OpenAPIResponse;
import com.wechi.adweb.bridge.exception.BadRequestException;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.analytics.dto.CreatePropertyRequestDTO;
import com.wechi.adweb.bridge.google.analytics.dto.GAAccountDTO;
import com.wechi.adweb.bridge.google.analytics.dto.GAPropertyDTO;
import com.wechi.adweb.bridge.google.analytics.dto.report.GAReportRequestDTO;
import com.wechi.adweb.bridge.google.analytics.dto.report.ReportType;
import com.wechi.adweb.bridge.google.analytics.dto.report.data.GAReportDataDTO;
import com.wechi.adweb.bridge.google.analytics.service.GAAdminService;
import com.wechi.adweb.bridge.google.analytics.service.GADataService;
import com.wechi.adweb.bridge.util.JsonUtils;

import io.swagger.v3.oas.annotations.Operation;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.util.CollectionUtils;
import org.springframework.web.bind.annotation.*;

import java.util.List;
import java.util.Objects;

/**
 * @author wfansh
 */
@Slf4j
@RestController
@RequestMapping("/api/google/ga")
public class GAController extends BaseController {
    @Autowired private GADataService gaDataService;

    @Autowired private GAAdminService gaAdminService;

    @PostMapping("/accounts/list")
    @ResponseBody
    public OpenAPIResponse<List<GAAccountDTO>> listAccounts() throws DataException {
        long start = System.currentTimeMillis();
        log.info("****** listAccounts() ******");

        List<GAAccountDTO> gaAccounts = gaAdminService.listAccounts();

        log.info("****** listAccounts() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<GAAccountDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(gaAccounts)
                .build();
    }

    @Operation(summary = "List the properties under a GA account resource name.")
    @PostMapping("/properties/list")
    @ResponseBody
    public OpenAPIResponse<List<GAPropertyDTO>> listProperties(
            @RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** listProperties() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String accountResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(accountResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        List<GAPropertyDTO> gaProperties =
                gaAdminService.listProperties(
                        // Converts to resource name if the account ID is provided by mistake.
                        GAAccountDTO.toResourceName(accountResourceName), false);
        log.info("****** listProperties() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<GAPropertyDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(gaProperties)
                .build();
    }

    @Operation(summary = "Get a property by its resource name.")
    @PostMapping("/properties/get")
    @ResponseBody
    public OpenAPIResponse<GAPropertyDTO> getProperty(
            @RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getProperty() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String propertyResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(propertyResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        GAPropertyDTO gaProperty =
                gaAdminService.getProperty(
                        // Converts to resource name if the property ID is provided by mistake.
                        GAPropertyDTO.toResourceName(propertyResourceName));
        log.info("****** getProperty() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<GAPropertyDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(gaProperty)
                .build();
    }

    @PostMapping("/properties/create")
    @ResponseBody
    public OpenAPIResponse<GAPropertyDTO> createPropertyWithDataStream(
            @RequestBody OpenAPIRequest<CreatePropertyRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info(
                "****** createPropertyWithDataStream() ****** apiRequest = {}",
                JsonUtils.toJson(apiRequest));
        CreatePropertyRequestDTO createPropertyRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(createPropertyRequest.getAccountResourceName())
                || StringUtils.isEmpty(createPropertyRequest.getDisplayName())
                || StringUtils.isEmpty(createPropertyRequest.getUrl())) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        GAPropertyDTO gaProperty =
                gaAdminService.createPropertyWithDataStream(
                        createPropertyRequest.getAccountResourceName(),
                        createPropertyRequest.getDisplayName(),
                        createPropertyRequest.getUrl());
        log.info(
                "****** createPropertyWithDataStream() ****** duration = {} seconds",
                getElapsedSeconds(start));
        return OpenAPIResponse.<GAPropertyDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(gaProperty)
                .build();
    }

    @Operation(summary = "Delete a property by its resource name.")
    @PostMapping("/properties/delete")
    @ResponseBody
    public OpenAPIResponse<String> deleteProperty(@RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** deleteProperty() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String propertyResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(propertyResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        gaAdminService.deleteProperty(
                // Converts to resource name if the property ID is provided by mistake.
                GAPropertyDTO.toResourceName(propertyResourceName));
        log.info("****** deleteProperty() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<String>builder()
                .status(APIStatus.SUCCESS)
                .data(propertyResourceName)
                .build();
    }

    @PostMapping("/report")
    @ResponseBody
    public OpenAPIResponse<List<? extends GAReportDataDTO>> runReport(
            @RequestBody OpenAPIRequest<GAReportRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** runReport() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        GAReportRequestDTO reportRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (Objects.isNull(reportRequest)
                || StringUtils.isEmpty(reportRequest.getPropertyResourceName())
                || Objects.isNull(reportRequest.getReportType())
                || StringUtils.isEmpty(reportRequest.getStartDate())
                || StringUtils.isEmpty(reportRequest.getEndDate())
                || (ReportType.ADWEB_CUSTOM_REPORT.equals(reportRequest.getReportType())
                        && CollectionUtils.isEmpty(reportRequest.getMetrics())
                        && CollectionUtils.isEmpty(reportRequest.getDimensions()))) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        List<? extends GAReportDataDTO> report = gaDataService.runReport(apiRequest.getData());
        log.info("****** runReport() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<? extends GAReportDataDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(report)
                .build();
    }
}
