package com.wechi.adweb.bridge.google.analytics.service;

import static com.wechi.adweb.bridge.google.common.ResourceDTO.parseId;

import com.google.analytics.admin.v1beta.*;
import com.google.api.client.util.Lists;
import com.google.api.gax.core.FixedCredentialsProvider;
import com.google.auth.oauth2.GoogleCredentials;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.analytics.dto.GAAccountDTO;
import com.wechi.adweb.bridge.google.analytics.dto.GADataStreamDTO;
import com.wechi.adweb.bridge.google.analytics.dto.GAPropertyDTO;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;

/**
 * @author wfansh
 */
@Slf4j
@Service
public class GAAdminService {

    private static final String DEFAULT_TIMEZONE = "Asia/Shanghai";

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    private AnalyticsAdminServiceSettings adminServiceSettings;

    @PostConstruct
    private void init() throws IOException {
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                        this.getClass().getClassLoader().getResourceAsStream(serviceAccountKey));

        this.adminServiceSettings =
                AnalyticsAdminServiceSettings.newBuilder()
                        .setCredentialsProvider(FixedCredentialsProvider.create(credentials))
                        .build();
    }

    public List<GAAccountDTO> listAccounts() throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            List<GAAccountDTO> gaAccounts = Lists.newArrayList();

            // Initializes the list request with default settings.
            ListAccountsRequest request = ListAccountsRequest.getDefaultInstance();
            for (Account account : analyticsAdminServiceClient.listAccounts(request).iterateAll()) {
                GAAccountDTO gaAccount =
                        GAAccountDTO.builder()
                                .id(parseId(account.getName()))
                                .resourceName(account.getName())
                                .displayName(account.getDisplayName())
                                .createTime(account.getCreateTime().getSeconds())
                                .updateTime(account.getUpdateTime().getSeconds())
                                .build();

                // Adds into the list;
                gaAccounts.add(gaAccount);
            }

            log.info("listAccounts : {}", JsonUtils.toJson(gaAccounts));
            return gaAccounts;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public List<GAPropertyDTO> listProperties(String accountResourceName, boolean withDataStreams)
            throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            List<GAPropertyDTO> gaProperties = Lists.newArrayList();

            // Initializes the list request with account resource name filter.
            // See
            // https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/properties/list#http-request
            ListPropertiesRequest request =
                    ListPropertiesRequest.newBuilder()
                            .setFilter("ancestor:" + accountResourceName)
                            .build();
            for (Property property :
                    analyticsAdminServiceClient.listProperties(request).iterateAll()) {
                GAPropertyDTO gaProperty =
                        GAPropertyDTO.builder()
                                .id(parseId(property.getName()))
                                .resourceName(property.getName())
                                .displayName(property.getDisplayName())
                                .propertyType(property.getPropertyType().toString())
                                .accountResourceName(property.getAccount())
                                .parentResourceName(property.getParent())
                                .industryCategory(property.getIndustryCategory().toString())
                                .timeZone(property.getTimeZone())
                                .currencyCode(property.getCurrencyCode())
                                .serviceLevel(property.getServiceLevel().toString())
                                .createTime(property.getCreateTime().getSeconds())
                                .updateTime(property.getUpdateTime().getSeconds())
                                // Sets the data streams associated with this property.
                                .dataStreams(
                                        withDataStreams
                                                ? listDataStreams(property.getName())
                                                : Collections.EMPTY_LIST)
                                .build();

                // Adds into the list;
                gaProperties.add(gaProperty);
            }

            // Distinct by property resource name.
            gaProperties =
                    gaProperties.stream()
                            .collect(Collectors.groupingBy(GAPropertyDTO::getResourceName))
                            .values()
                            .stream()
                            .map(values -> values.get(0))
                            .toList();

            log.info(
                    "listProperties for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gaProperties));
            return gaProperties;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    private List<GADataStreamDTO> listDataStreams(String propertyResourceName)
            throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            List<GADataStreamDTO> gaDataStreams = Lists.newArrayList();

            // Initializes the list request with parent property.
            ListDataStreamsRequest request =
                    ListDataStreamsRequest.newBuilder().setParent(propertyResourceName).build();
            for (DataStream dataStream :
                    analyticsAdminServiceClient.listDataStreams(request).iterateAll()) {
                GADataStreamDTO gaDataStream =
                        GADataStreamDTO.builder()
                                .id(parseId(dataStream.getName()))
                                .resourceName(dataStream.getName())
                                .displayName(dataStream.getDisplayName())
                                .propertyResourceName(propertyResourceName)
                                .streamMeasurementId(
                                        dataStream.getWebStreamData().getMeasurementId())
                                .streamDefaultUrl(dataStream.getWebStreamData().getDefaultUri())
                                .build();

                // Adds into the list;
                gaDataStreams.add(gaDataStream);
            }

            log.info(
                    "listDataStreams for {} : {}",
                    propertyResourceName,
                    JsonUtils.toJson(gaDataStreams));
            return gaDataStreams;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GAPropertyDTO getProperty(String propertyResourceName) throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            Property property = analyticsAdminServiceClient.getProperty(propertyResourceName);

            // Converts and returns.
            GAPropertyDTO gaProperty =
                    GAPropertyDTO.builder()
                            .id(parseId(property.getName()))
                            .resourceName(property.getName())
                            .displayName(property.getDisplayName())
                            .propertyType(property.getPropertyType().toString())
                            .accountResourceName(property.getAccount())
                            .parentResourceName(property.getParent())
                            .industryCategory(property.getIndustryCategory().toString())
                            .timeZone(property.getTimeZone())
                            .currencyCode(property.getCurrencyCode())
                            .serviceLevel(property.getServiceLevel().toString())
                            .createTime(property.getCreateTime().getSeconds())
                            .updateTime(property.getUpdateTime().getSeconds())
                            // Sets the data streams associated with this property.
                            .dataStreams(listDataStreams(property.getName()))
                            .build();

            log.info("getProperty : {}", JsonUtils.toJson(gaProperty));
            return gaProperty;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GAPropertyDTO createPropertyWithDataStream(
            String accountResourceName, String displayName, String url) throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            // 1. Creates the property.
            Property property =
                    Property.newBuilder()
                            .setParent(accountResourceName)
                            .setDisplayName(displayName)
                            .setTimeZone(DEFAULT_TIMEZONE)
                            .build();
            Property createdProperty = analyticsAdminServiceClient.createProperty(property);

            // 2. Creates the data stream.
            DataStream dataStream =
                    DataStream.newBuilder()
                            .setDisplayName(displayName)
                            .setType(DataStream.DataStreamType.WEB_DATA_STREAM)
                            .setWebStreamData(
                                    DataStream.WebStreamData.newBuilder().setDefaultUri(url))
                            .build();
            DataStream createdDataStream =
                    analyticsAdminServiceClient.createDataStream(
                            createdProperty.getName(), dataStream);

            // 3. Converts and returns.
            GADataStreamDTO gaDataStream =
                    GADataStreamDTO.builder()
                            .id(parseId(createdDataStream.getName()))
                            .resourceName(createdDataStream.getName())
                            .displayName(createdDataStream.getDisplayName())
                            .propertyResourceName(createdProperty.getName())
                            .streamMeasurementId(
                                    createdDataStream.getWebStreamData().getMeasurementId())
                            .streamDefaultUrl(createdDataStream.getWebStreamData().getDefaultUri())
                            .build();

            GAPropertyDTO gaProperty =
                    GAPropertyDTO.builder()
                            .id(parseId(createdProperty.getName()))
                            .resourceName(createdProperty.getName())
                            .displayName(createdProperty.getDisplayName())
                            .propertyType(createdProperty.getPropertyType().toString())
                            .accountResourceName(createdProperty.getAccount())
                            .parentResourceName(createdProperty.getParent())
                            .industryCategory(createdProperty.getIndustryCategory().toString())
                            .timeZone(createdProperty.getTimeZone())
                            .currencyCode(createdProperty.getCurrencyCode())
                            .serviceLevel(createdProperty.getServiceLevel().toString())
                            .createTime(createdProperty.getCreateTime().getSeconds())
                            .updateTime(createdProperty.getUpdateTime().getSeconds())
                            .dataStreams(Collections.singletonList(gaDataStream))
                            .build();

            log.info(
                    "createPropertyWithDataStream for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gaProperty));
            return gaProperty;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public void deleteProperty(String propertyResourceName) throws DataException {
        try (AnalyticsAdminServiceClient analyticsAdminServiceClient =
                AnalyticsAdminServiceClient.create(adminServiceSettings)) {
            analyticsAdminServiceClient.deleteProperty(propertyResourceName);
            log.info("deleteProperty {}", propertyResourceName);
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }
}
