package com.wechi.adweb.bridge.google.analytics.service;

import com.google.analytics.data.v1beta.*;
import com.google.api.gax.core.FixedCredentialsProvider;
import com.google.auth.oauth2.GoogleCredentials;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.analytics.dto.report.GAReportRequestDTO;
import com.wechi.adweb.bridge.google.analytics.dto.report.OrderByType;
import com.wechi.adweb.bridge.google.analytics.dto.report.ReportType;
import com.wechi.adweb.bridge.google.analytics.dto.report.data.*;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import javax.annotation.PostConstruct;

/**
 * @author wfansh
 *     <p>See https://developers.google.com/analytics/devguides/reporting/data/v1/api-schema for GA
 *     metrics and dimensions.
 */
@Slf4j
@Service
public class GADataService {

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    private BetaAnalyticsDataSettings dataSettings;

    @PostConstruct
    private void init() throws IOException {
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                        this.getClass().getClassLoader().getResourceAsStream(serviceAccountKey));

        this.dataSettings =
                BetaAnalyticsDataSettings.newBuilder()
                        .setCredentialsProvider(FixedCredentialsProvider.create(credentials))
                        .build();
    }

    public List<? extends GAReportDataDTO> runReport(GAReportRequestDTO reportRequest)
            throws DataException {
        // 0. Enhances the report request with default settings.
        this.amplifyReportRequest(reportRequest);

        try (BetaAnalyticsDataClient analyticsDataClient =
                BetaAnalyticsDataClient.create(dataSettings)) {
            // 1. Dimensions.
            List<Dimension> dimensions =
                    reportRequest.getDimensions().stream()
                            .map(String::trim)
                            .filter(StringUtils::isNotEmpty)
                            .map(dimension -> Dimension.newBuilder().setName(dimension).build())
                            .toList();

            // 2. Metrics.
            List<Metric> metrics =
                    reportRequest.getMetrics().stream()
                            .map(String::trim)
                            .filter(StringUtils::isNotEmpty)
                            .map(metric -> Metric.newBuilder().setName(metric).build())
                            .toList();

            // 3. Date range.
            DateRange dateRange =
                    DateRange.newBuilder()
                            .setStartDate(reportRequest.getStartDate())
                            .setEndDate(reportRequest.getEndDate())
                            .build();

            // 4. Order by.
            OrderBy.Builder orderBy = null;
            String orderByName = reportRequest.getOrderBy();
            OrderByType orderByType = reportRequest.getOrderByType();
            if (StringUtils.isNotEmpty(orderByName) && Objects.nonNull(orderByType)) {
                orderBy = OrderBy.newBuilder();
                switch (orderByType) {
                    case DIMENSIONS ->
                            orderBy.setDimension(
                                    OrderBy.DimensionOrderBy.newBuilder()
                                            .setDimensionName(orderByName));
                    case METRICS ->
                            orderBy.setMetric(
                                    OrderBy.MetricOrderBy.newBuilder().setMetricName(orderByName));
                }

                // Desc.
                orderBy.setDesc(BooleanUtils.isTrue(reportRequest.getOrderByDesc()));
            }

            RunReportRequest.Builder request =
                    RunReportRequest.newBuilder()
                            .setProperty(reportRequest.getPropertyResourceName())
                            .addAllDimensions(dimensions)
                            .addAllMetrics(metrics)
                            .addDateRanges(dateRange)
                            // The API returns a maximum of 250,000 rows per request. See
                            // https://developers.google.com/analytics/devguides/reporting/data/v1/rest/v1beta/properties/runReport
                            .setLimit(250_000L);
            if (Objects.nonNull(orderBy)) {
                request.addOrderBys(orderBy);
            }

            RunReportResponse response = analyticsDataClient.runReport(request.build());

            // Converts and returns.
            List<? extends GAReportDataDTO> report =
                    toReport(response.getRowsList(), reportRequest);
            log.info(
                    "Report type = {} - {}",
                    reportRequest.getReportType(),
                    JsonUtils.toJson(report));
            return report;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    /** Enhances the report request with default settings from {@link ReportType}. */
    private void amplifyReportRequest(GAReportRequestDTO reportRequest) {
        ReportType reportType = reportRequest.getReportType();
        if (Objects.isNull(reportType)) {
            return;
        }

        if (CollectionUtils.isEmpty(reportRequest.getMetrics())) {
            reportRequest.setMetrics(reportType.getDefaultMetrics());
        }

        if (CollectionUtils.isEmpty(reportRequest.getDimensions())) {
            reportRequest.setDimensions(reportType.getDefaultDimensions());
        }

        if (Objects.isNull(reportRequest.getOrderBy())) {
            reportRequest.setOrderBy(reportType.getDefaultOrderBy());
        }

        if (Objects.isNull(reportRequest.getOrderByType())) {
            reportRequest.setOrderByType(reportType.getDefaultOrderByType());
        }

        if (Objects.isNull(reportRequest.getOrderByDesc())) {
            reportRequest.setOrderByDesc(reportType.getDefaultOrderByDesc());
        }
    }

    private List<? extends GAReportDataDTO> toReport(
            List<Row> reportRows, GAReportRequestDTO reportRequest) {
        if (Objects.isNull(reportRequest.getReportType())) {
            return Collections.EMPTY_LIST;
        }

        List<String> metrics = reportRequest.getMetrics();
        List<String> dimensions = reportRequest.getDimensions();

        return switch (reportRequest.getReportType()) {
            case ADWEB_COUNTRY_CHART -> CountryChartData.toReport(reportRows, metrics, dimensions);
            case ADWEB_SESSION_SOURCE_MEDIUM_VIEW ->
                    SourceMediaViewData.toReport(reportRows, metrics, dimensions);
            case ADWEB_PAGE_PATH_VIEW -> PagePathViewData.toReport(reportRows, metrics, dimensions);
            case ADWEB_DATE_VIEW -> DateViewData.toReport(reportRows, metrics, dimensions);
            case ADWEB_CUSTOM_REPORT -> CustomReportData.toReport(reportRows, metrics, dimensions);
        };
    }
}
