package com.wechi.adweb.bridge.google.gtm.controller;

import com.wechi.adweb.bridge.common.APIStatus;
import com.wechi.adweb.bridge.common.BaseController;
import com.wechi.adweb.bridge.common.OpenAPIRequest;
import com.wechi.adweb.bridge.common.OpenAPIResponse;
import com.wechi.adweb.bridge.exception.BadRequestException;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.gtm.dto.CreateContainerRequestDTO;
import com.wechi.adweb.bridge.google.gtm.dto.GTMAccountDTO;
import com.wechi.adweb.bridge.google.gtm.dto.GTMContainerDTO;
import com.wechi.adweb.bridge.google.gtm.service.GTMService;
import com.wechi.adweb.bridge.util.JsonUtils;

import io.swagger.v3.oas.annotations.Operation;

import lombok.extern.slf4j.Slf4j;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.List;

/**
 * @author wfansh
 */
@Slf4j
@RestController
@RequestMapping("/api/google/gtm")
public class GTMController extends BaseController {

    @Autowired private GTMService gtmService;

    @PostMapping("/accounts/list")
    @ResponseBody
    public OpenAPIResponse<List<GTMAccountDTO>> listAccounts() throws DataException {
        long start = System.currentTimeMillis();
        log.info("****** listAccounts() ******");

        List<GTMAccountDTO> gtmAccounts = gtmService.listAccounts();

        log.info("****** listAccounts() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<GTMAccountDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(gtmAccounts)
                .build();
    }

    @Operation(summary = "List the containers under a GTM account resource name.")
    @PostMapping("/containers/list")
    @ResponseBody
    public OpenAPIResponse<List<GTMContainerDTO>> listContainers(
            @RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** listContainers() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String accountResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(accountResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        List<GTMContainerDTO> gtmContainers =
                gtmService.listContainers(
                        // Converts to resource name if the account ID is provided by mistake.
                        GTMAccountDTO.toResourceName(accountResourceName));
        log.info("****** listContainers() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<List<GTMContainerDTO>>builder()
                .status(APIStatus.SUCCESS)
                .data(gtmContainers)
                .build();
    }

    @Operation(summary = "Get a container by its resource name.")
    @PostMapping("/containers/get")
    @ResponseBody
    public OpenAPIResponse<GTMContainerDTO> getContainer(
            @RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** getContainer() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String containerResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(containerResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        GTMContainerDTO gtmContainer = gtmService.getContainer(containerResourceName);
        log.info("****** getContainer() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<GTMContainerDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(gtmContainer)
                .build();
    }

    @PostMapping("/containers/create")
    @ResponseBody
    public OpenAPIResponse<GTMContainerDTO> createContainer(
            @RequestBody OpenAPIRequest<CreateContainerRequestDTO> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** createContainer() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        CreateContainerRequestDTO createContainerRequest = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(createContainerRequest.getAccountResourceName())
                || StringUtils.isEmpty(createContainerRequest.getDisplayName())
                || StringUtils.isEmpty(createContainerRequest.getGoogleTagId())) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        GTMContainerDTO gtmContainer =
                gtmService.createContainer(
                        createContainerRequest.getAccountResourceName(),
                        createContainerRequest.getDisplayName());

        // 3. Initializes the container in ASYNC mode.
        gtmService.initContainerAccess(
                createContainerRequest.getAccountResourceName(), gtmContainer.getId());
        gtmService.initContainerSetups(
                gtmContainer.getResourceName(), createContainerRequest.getGoogleTagId());

        log.info("****** createContainer() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<GTMContainerDTO>builder()
                .status(APIStatus.SUCCESS)
                .data(gtmContainer)
                .build();
    }

    @Operation(summary = "Delete a container by its resource name.")
    @PostMapping("/containers/delete")
    @ResponseBody
    public OpenAPIResponse<String> deleteContainer(@RequestBody OpenAPIRequest<String> apiRequest)
            throws BadRequestException, DataException {
        long start = System.currentTimeMillis();
        log.info("****** deleteContainer() ****** apiRequest = {}", JsonUtils.toJson(apiRequest));
        String containerResourceName = apiRequest.getData();

        // 1. Validates the request parameters.
        if (StringUtils.isEmpty(containerResourceName)) {
            throw new BadRequestException(apiRequest);
        }

        // 2. Executes the API request.
        gtmService.deleteContainer(containerResourceName);
        log.info("****** deleteContainer() ****** duration = {} seconds", getElapsedSeconds(start));
        return OpenAPIResponse.<String>builder()
                .status(APIStatus.SUCCESS)
                .data(containerResourceName)
                .build();
    }
}
